@@ -2,12 +2,26 @@
  * the slang SLtt interface.
  */
 
-/* Copyright (c) 1992, 1999, 2001 John E. Davis
- * This file is part of the S-Lang library.
- *
- * You may distribute under the terms of either the GNU General Public
- * License or the Perl Artistic License.
- */
+/*
+Copyright (C) 2004, 2005, 2006, 2007 John E. Davis
+
+This file is part of the S-Lang Library.
+
+The S-Lang Library is free software; you can redistribute it and/or
+modify it under the terms of the GNU General Public License as
+published by the Free Software Foundation; either version 2 of the
+License, or (at your option) any later version.
+
+The S-Lang Library is distributed in the hope that it will be useful,
+but WITHOUT ANY WARRANTY; without even the implied warranty of
+MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
+General Public License for more details.
+
+You should have received a copy of the GNU General Public License
+along with this library; if not, write to the Free Software
+Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
+USA.  
+*/
 
 #include "slinclud.h"
 
@@ -63,7 +77,7 @@ static int make_integer (unsigned char *buf)
  * structure is SLTERMCAP.  In that case, only terminal_names is malloced
  * and the other fields are pointers into it.
  */
-struct _SLterminfo_Type
+struct _pSLterminfo_Type
 {
 #define SLTERMINFO 1
 #define SLTERMCAP  2
@@ -216,8 +230,7 @@ static char *read_string_table (FILE *fp, SLterminfo_Type *t)
  * are implemented by multiple links to the same compiled file.
  */
 
-#define MAX_TI_DIRS 7
-static char *Terminfo_Dirs [MAX_TI_DIRS] =
+static char *Terminfo_Dirs [] =
 {
    NULL, /* $HOME/.terminfo */
    NULL, /* $TERMINFO */
@@ -225,10 +238,27 @@ static char *Terminfo_Dirs [MAX_TI_DIRS] =
    "/usr/lib/terminfo",
    "/usr/share/lib/terminfo",
    "/etc/terminfo",
-   "/usr/local/lib/terminfo"
+   "/usr/local/lib/terminfo",
+#ifdef MISC_TERMINFO_DIRS
+   MISC_TERMINFO_DIRS,
+#endif
+   ""
 };
 
-SLterminfo_Type *_SLtt_tigetent (char *term)
+void _pSLtt_tifreeent (SLterminfo_Type *t)
+{
+   if (t == NULL)
+     return;
+
+   SLfree ((char *)t->string_table);
+   SLfree ((char *)t->string_offsets);
+   SLfree ((char *)t->numbers);
+   SLfree ((char *)t->boolean_flags);
+   SLfree ((char *)t->terminal_names);
+   SLfree ((char *)t);
+}
+
+SLterminfo_Type *_pSLtt_tigetent (char *term)
 {
    char *tidir;
    int i;
@@ -246,10 +276,15 @@ SLterminfo_Type *_SLtt_tigetent (char *term)
        )
      return NULL;
 
+   if (_pSLsecure_issetugid ()
+       && ((term[0] == '.') || (NULL != strchr (term, '/'))))
+     return NULL;
+
    if (NULL == (ti = (SLterminfo_Type *) SLmalloc (sizeof (SLterminfo_Type))))
      {
 	return NULL;
      }
+   memset ((char *)ti, 0, sizeof (SLterminfo_Type));
 
 #ifdef SLANG_UNTIC
    if (SLang_Untic_Terminfo_File != NULL)
@@ -262,7 +297,7 @@ SLterminfo_Type *_SLtt_tigetent (char *term)
    /* If we are on a termcap based system, use termcap */
    if (0 == tcap_getent (term, ti)) return ti;
 
-   if (NULL != (home = getenv ("HOME")))
+   if (NULL != (home = _pSLsecure_getenv ("HOME")))
      {
 	strncpy (home_ti, home, sizeof (home_ti) - 11);
 	home_ti [sizeof(home_ti) - 11] = 0;
@@ -270,54 +305,48 @@ SLterminfo_Type *_SLtt_tigetent (char *term)
 	Terminfo_Dirs [0] = home_ti;
      }
 
-   Terminfo_Dirs[1] = getenv ("TERMINFO");
+   Terminfo_Dirs[1] = _pSLsecure_getenv ("TERMINFO");
    i = 0;
-   while (i < MAX_TI_DIRS)
+   while (1)
      {
 	tidir = Terminfo_Dirs[i];
-	if ((tidir != NULL)
-	    && (sizeof (file) > strlen (tidir) + 2 + strlen (term)))
+	if (tidir != NULL)
 	  {
-	     sprintf (file, "%s/%c/%s", tidir, *term, term);
-	     if (NULL != (fp = open_terminfo (file, ti)))
-	       break;
+	     if (*tidir == 0)
+	       break;		       /* last one */
+
+	     if (sizeof (file) > strlen (tidir) + 4 + strlen (term))
+	       {
+		  sprintf (file, "%s/%c/%s", tidir, *term, term);
+		  if (NULL != (fp = open_terminfo (file, ti)))
+		    break;
+	       }
 	  }
 	i++;
      }
+
 #ifdef SLANG_UNTIC
    fp_open_label:
 #endif
 
-   if (fp != NULL)
+   if (fp == NULL)
      {
-	if (NULL != read_terminal_names (fp, ti))
-	  {
-	     if (NULL != read_boolean_flags (fp, ti))
-	       {
-		  if (NULL != read_numbers (fp, ti))
-		    {
-		       if (NULL != read_string_offsets (fp, ti))
-			 {
-			    if (NULL != read_string_table (fp, ti))
-			      {
-				 /* success */
-				 fclose (fp);
-				 ti->flags = SLTERMINFO;
-				 return ti;
-			      }
-			    SLfree ((char *)ti->string_offsets);
-			 }
-		       SLfree ((char *)ti->numbers);
-		    }
-		  SLfree ((char *)ti->boolean_flags);
-	       }
-	     SLfree ((char *)ti->terminal_names);
-	  }
-	fclose (fp);
+	SLfree ((char *) ti);
+	return NULL;
      }
 
-   SLfree ((char *)ti);
-   return NULL;
+   ti->flags = SLTERMINFO;
+   if ((NULL == read_terminal_names (fp, ti))
+       || (NULL == read_boolean_flags (fp, ti))
+       || (NULL == read_numbers (fp, ti))
+       || (NULL == read_string_offsets (fp, ti))
+       || (NULL == read_string_table (fp, ti)))
+     {
+	_pSLtt_tifreeent (ti);
+	ti = NULL;
+     }
+   (void) fclose (fp);
+   return ti;
 }
 
 #ifdef SLANG_UNTIC
@@ -326,7 +355,7 @@ SLterminfo_Type *_SLtt_tigetent (char *term)
 # define UNTIC_COMMENT(x)
 #endif
 
-typedef struct
+typedef SLCONST struct
 {
    char name[3];
    int offset;
@@ -773,7 +802,7 @@ static int compute_cap_offset (char *cap, SLterminfo_Type *t, Tgetstr_Map_Type *
    return -1;
 }
 
-char *_SLtt_tigetstr (SLterminfo_Type *t, char *cap)
+char *_pSLtt_tigetstr (SLterminfo_Type *t, char *cap)
 {
    int offset;
 
@@ -833,7 +862,7 @@ static Tgetstr_Map_Type Tgetnum_Map[] =
    {"", -1		UNTIC_COMMENT(NULL)}
 };
 
-int _SLtt_tigetnum (SLterminfo_Type *t, char *cap)
+int _pSLtt_tigetnum (SLterminfo_Type *t, char *cap)
 {
    int offset;
 
@@ -896,7 +925,7 @@ static Tgetstr_Map_Type Tgetflag_Map[] =
    {"", -1		UNTIC_COMMENT(NULL)}
 };
 
-int _SLtt_tigetflag (SLterminfo_Type *t, char *cap)
+int _pSLtt_tigetflag (SLterminfo_Type *t, char *cap)
 {
    int offset;
 
@@ -994,6 +1023,16 @@ static int tcap_getent (char *term, SLterminfo_Type *ti)
    termcap = (unsigned char *) getenv ("TERMCAP");
    if ((termcap == NULL) || (*termcap == '/')) return -1;
 
+   /* SUN Solaris 7&8 have bug in tset program under tcsh,
+    * eval `tset -s -A -Q` sets value of TERMCAP to ":",
+    *  under other shells it works fine.
+    *  SUN was informed, they marked it as duplicate of bug 4086585
+    *  but didn't care to fix it... <mikkopa@cs.tut.fi> 
+    */
+   if ((termcap[0] == ':') && (termcap[1] == 0))
+     return -1;
+   
+   
    /* We have a termcap so lets use it provided it does not have a reference
     * to another terminal via tc=.  In that case, use terminfo.  The alternative
     * would be to parse the termcap file which I do not want to do right now.
@@ -1011,7 +1050,8 @@ static int tcap_getent (char *term, SLterminfo_Type *ti)
 
    /* malloc some extra space just in case it is needed. */
    len = strlen ((char *) termcap) + 256;
-   if (NULL == (buf = (unsigned char *) SLmalloc ((unsigned int) len))) return -1;
+   if (NULL == (buf = (unsigned char *) SLmalloc ((unsigned int) len))) 
+     return -1;
 
    b = buf;
 
@@ -1059,7 +1099,15 @@ static int tcap_getent (char *term, SLterminfo_Type *ti)
 	     ch = *t++;
 	     if ((ch == '\\') && (t < tmax))
 	       {
-		  t = (unsigned char *) _SLexpand_escaped_char ((char *) t, (char *) &ch);
+		  SLwchar_Type wch;
+
+		  t = (unsigned char *) _pSLexpand_escaped_char ((char *) t, &wch, NULL);
+		  if (t == NULL)
+		    {
+		       SLfree ((char *)buf);
+		       return -1;
+		    }
+		  ch = (char) wch;
 	       }
 	     else if ((ch == '^') && (t < tmax))
 	       {
@@ -1135,21 +1183,21 @@ static int tcap_getent (char *term, SLterminfo_Type *ti)
  */
 char *SLtt_tigetent (char *s)
 {
-   return (char *) _SLtt_tigetent (s);
+   return (char *) _pSLtt_tigetent (s);
 }
 
 extern char *SLtt_tigetstr (char *s, char **p)
 {
    if (p == NULL)
      return NULL;
-   return _SLtt_tigetstr ((SLterminfo_Type *) *p, s);
+   return _pSLtt_tigetstr ((SLterminfo_Type *) *p, s);
 }
 
 extern int SLtt_tigetnum (char *s, char **p)
 {
    if (p == NULL)
      return -1;
-   return _SLtt_tigetnum ((SLterminfo_Type *) *p, s);
+   return _pSLtt_tigetnum ((SLterminfo_Type *) *p, s);
 }
 
 
